import bisect
from typing import Iterable, Sequence

from rainbowadn.hashing.hash_point_format import hash_point_format, tabulate
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullable import Nullable
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable

__all__ = ('WeakReferenceIndexSetBTree', 'WrisbtFactory',)


class WeakReferenceIndexSetBTree(RecursiveMentionable):
    def __init__(
            self,
            source: bytes,
            height: int,
            keymin: int,
            root: bool,
            cache: tuple[Nullable['WeakReferenceIndexSetBTree'], ...]
    ):
        assert isinstance(source, bytes)
        assert isinstance(height, int)
        assert isinstance(keymin, int)
        assert isinstance(root, bool)
        assert isinstance(cache, tuple)

        self.length = len(source) // HashPoint.HASH_LENGTH
        assert len(source) == self.length * HashPoint.HASH_LENGTH
        self.source = source

        assert height >= 0
        self.height = height
        self.leaf = height == 0

        self.empty = self.length == 0
        if self.empty:
            assert self.leaf

        if self.leaf:
            self.keys = self.length
            self.children = 0
        else:
            self.keys = self.length // 2
            self.children = self.keys + 1
        assert self.length == self.keys + self.children

        assert keymin > 0
        self.keymin = keymin

        self.root = root

        if not root:
            assert keymin <= self.keys
        assert self.keys <= 2 * keymin + 1

        assert len(cache) == self.children
        child_index: int
        cached: Nullable[WeakReferenceIndexSetBTree]
        self.cache = cache

        if self.balanced():
            self._hash_point = HashPoint.of(self)

    def hash_point(self) -> HashPoint['WeakReferenceIndexSetBTree']:
        return self._hash_point

    def full(self) -> bool:
        return self.keys == 2 * self.keymin + 1

    def bytes_no(self, index: int) -> bytes:
        assert isinstance(index, int)
        assert 0 <= index < self.length
        return self.source[HashPoint.HASH_LENGTH * index:HashPoint.HASH_LENGTH * (index + 1)]

    def key_no(self, index: int) -> bytes:
        assert isinstance(index, int)
        assert 0 <= index < self.keys
        return self.bytes_no(index)

    def cached_no(self, index: int) -> Nullable['WeakReferenceIndexSetBTree']:
        assert isinstance(index, int)
        assert 0 <= index < self.children
        assert not self.leaf
        cached: Nullable[WeakReferenceIndexSetBTree] = self.cache[index]
        assert isinstance(cached, Nullable)
        if isinstance(cached, NotNull):
            tree: WeakReferenceIndexSetBTree = cached.value
            assert self.bytes_no(self.keys + index) == tree.hash_point().point
        return cached

    def child_no(self, index: int) -> HashPoint['WeakReferenceIndexSetBTree']:
        assert isinstance(index, int)
        assert 0 <= index < self.children
        assert not self.leaf
        return HashPoint(
            WrisbtFactory(self.height - 1, self.keymin, False),
            self.bytes_no(self.keys + index),
            self.cached_no(index)
        )

    def balanced(self) -> bool:
        return self.keys <= 2 * self.keymin

    def points(self) -> Iterable[HashPoint]:
        assert self.balanced()
        if self.leaf:
            return []
        else:
            return [self.child_no(child_index) for child_index in range(self.children)]

    def __bytes__(self):
        assert self.balanced()
        return self.source

    def __topology_hash__(self) -> bytes:
        return HashPoint.hash(self.source[self.keys * HashPoint.HASH_LENGTH:])

    def __factory__(self) -> RainbowFactory['WeakReferenceIndexSetBTree']:
        assert self.balanced()
        return WrisbtFactory(self.height, self.keymin, self.root)

    def range(self, left: int, right: int) -> 'WeakReferenceIndexSetBTree':
        assert isinstance(left, int)
        assert isinstance(right, int)
        assert 0 <= left < right <= self.keys
        hl: int = HashPoint.HASH_LENGTH
        assert isinstance(hl, int)
        if self.leaf:
            return WeakReferenceIndexSetBTree(
                self.source[hl * left:hl * right],
                0,
                self.keymin,
                False,
                ()
            )
        else:
            keybytes: bytes = self.source[hl * left:hl * right]
            assert isinstance(keybytes, bytes)
            childbytes: bytes = self.source[hl * (self.keys + left):hl * (self.keys + right + 1)]
            assert isinstance(childbytes, bytes)
            return WeakReferenceIndexSetBTree(
                keybytes + childbytes,
                self.height,
                self.keymin,
                False,
                self.cache[left:right + 1],
            )

    def split(self) -> tuple['WeakReferenceIndexSetBTree', bytes, 'WeakReferenceIndexSetBTree']:
        assert self.full()
        return (
            self.range(0, self.keymin),
            self.key_no(self.keymin),
            self.range(self.keymin + 1, 2 * self.keymin + 1),
        )

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)

        formatted = f'{self.height}' \
                    f'{tabulate(tab)}('
        for key_index in range(self.keys):
            formatted += f'{tabulate(tab + 1)}{self.key_no(key_index).hex()}'
        for child_index in range(self.children):
            formatted += f'{tabulate(tab + 1)}{hash_point_format(self.child_no(child_index), resolver, tab + 1)}'

        return f'{formatted}' \
               f'{tabulate(tab)})'

    def contains(self, resolver: HashResolver, key: bytes) -> bool:
        assert isinstance(resolver, HashResolver)
        assert isinstance(key, bytes)
        assert len(key) == HashPoint.HASH_LENGTH
        hl: int = HashPoint.HASH_LENGTH
        assert isinstance(hl, int)

        assert self.balanced()

        index: int = bisect.bisect_left(KeyView(self), key)
        assert isinstance(index, int)

        if index < self.keys and self.key_no(index) == key:
            return True
        if index < self.keys:
            assert key < self.key_no(index)
        if index > 0:
            assert key > self.key_no(index - 1)
        if self.leaf:
            return False
        child: WeakReferenceIndexSetBTree = resolver.resolve(self.child_no(index))
        assert isinstance(child, WeakReferenceIndexSetBTree)
        return child.contains(resolver, key)

    def add(self, resolver: HashResolver, key: bytes) -> 'WeakReferenceIndexSetBTree':
        assert isinstance(resolver, HashResolver)
        assert isinstance(key, bytes)
        assert len(key) == HashPoint.HASH_LENGTH
        hl: int = HashPoint.HASH_LENGTH
        assert isinstance(hl, int)

        assert self.balanced()

        index: int = bisect.bisect_left(KeyView(self), key)
        assert isinstance(index, int)

        if index < self.keys and self.key_no(index) == key:
            return self
        if index < self.keys:
            assert key < self.key_no(index)
        if index > 0:
            assert key > self.key_no(index - 1)
        if self.leaf:
            return WeakReferenceIndexSetBTree(
                self.source[:hl * index] + key + self.source[hl * index:],
                self.height,
                self.keymin,
                self.root,
                ()
            )
        child: WeakReferenceIndexSetBTree = resolver.resolve(self.child_no(index))
        assert isinstance(child, WeakReferenceIndexSetBTree)
        child: WeakReferenceIndexSetBTree = child.add(resolver, key)
        assert isinstance(child, WeakReferenceIndexSetBTree)
        if child.full():
            left, middle, right = child.split()
            assert isinstance(left, WeakReferenceIndexSetBTree)
            assert isinstance(middle, bytes)
            assert isinstance(right, WeakReferenceIndexSetBTree)
            return WeakReferenceIndexSetBTree(
                (
                        self.source[:hl * index]
                        +
                        middle
                        +
                        self.source[hl * index:hl * (self.keys + index)]
                        +
                        bytes(left.hash_point())
                        +
                        bytes(right.hash_point())
                        +
                        self.source[hl * (self.keys + index + 1):]
                ),
                self.height,
                self.keymin,
                self.root,
                self.cache[:index] + (NotNull(left), NotNull(right)) + self.cache[index + 1:]
            )
        else:
            return WeakReferenceIndexSetBTree(
                (
                        self.source[:hl * (self.keys + index)]
                        +
                        bytes(HashPoint.of(child))
                        +
                        self.source[hl * (self.keys + index + 1):]
                ),
                self.height,
                self.keymin,
                self.root,
                self.cache[:index] + (NotNull(child),) + self.cache[index + 1:]
            )

    def iter_keys(self, resolver: HashResolver) -> Iterable[bytes]:
        if self.leaf:
            for key_index in range(self.keys):
                yield self.key_no(key_index)
        else:
            for index in range(self.length):
                real_index, mode = divmod(index, 2)
                if mode:
                    yield self.key_no(real_index)
                else:
                    yield from resolver.resolve(self.child_no(real_index)).iter_keys(resolver)


class KeyView(Sequence[WeakReferenceIndexSetBTree]):
    def __init__(self, wrisbt: WeakReferenceIndexSetBTree):
        assert isinstance(wrisbt, WeakReferenceIndexSetBTree)
        self.wrisbt = wrisbt

    def __len__(self):
        return self.wrisbt.keys

    def __getitem__(self, index: int) -> bytes:
        return self.wrisbt.key_no(index)


class WrisbtFactory(RainbowFactory[WeakReferenceIndexSetBTree]):
    def __init__(self, height: int, keymin: int, root: bool):
        assert isinstance(height, int)
        assert isinstance(keymin, int)
        assert isinstance(root, bool)
        assert height >= 0
        self.height = height
        assert keymin > 0
        self.keymin = keymin
        self.root = root

    def from_bytes(self, source: bytes) -> WeakReferenceIndexSetBTree:
        assert isinstance(source, bytes)
        return WeakReferenceIndexSetBTree(
            source,
            self.height,
            self.keymin,
            self.root,
            (Null(),) * (len(source) // (2 * HashPoint.HASH_LENGTH) + 1) if self.height else ()
        )
